@@ -1,6 +1,6 @@
 use crate::memory::{LinearMemory, MemoryCreator};
-use crate::trampoline::{create_handle, StoreInstanceHandle};
-use crate::Store;
+use crate::store::{InstanceId, StoreOpaque};
+use crate::trampoline::create_handle;
 use crate::{Limits, MemoryType};
 use anyhow::{anyhow, Result};
 use wasmtime_environ::entity::PrimaryMap;
@@ -9,7 +9,7 @@ use wasmtime_runtime::{RuntimeLinearMemory, RuntimeMemoryCreator, VMMemoryDefini
 
 use std::sync::Arc;
 
-pub fn create_memory(store: &Store, memory: &MemoryType) -> Result<StoreInstanceHandle> {
+pub fn create_memory(store: &mut StoreOpaque<'_>, memory: &MemoryType) -> Result<InstanceId> {
     let mut module = Module::new();
 
     let memory = wasm::Memory {
@@ -41,7 +41,7 @@ impl RuntimeLinearMemory for LinearMemoryProxy {
         self.mem.maximum()
     }
 
-    fn grow(&self, delta: u32) -> Option<u32> {
+    fn grow(&mut self, delta: u32) -> Option<u32> {
         self.mem.grow(delta)
     }
 
