@@ -1,19 +1,23 @@
-use crate::{signatures::SignatureCollection, Extern, Store};
+use crate::linker::Definition;
+use crate::store::StoreData;
+use crate::{signatures::SignatureCollection, Engine, Extern};
 use anyhow::{bail, Context, Result};
 use wasmtime_environ::wasm::{
     EntityType, Global, InstanceTypeIndex, Memory, ModuleTypeIndex, SignatureIndex, Table,
 };
 use wasmtime_jit::TypeTables;
+use wasmtime_runtime::VMSharedSignatureIndex;
 
 pub struct MatchCx<'a> {
     pub signatures: &'a SignatureCollection,
     pub types: &'a TypeTables,
-    pub store: &'a Store,
+    pub store_data: &'a StoreData,
+    pub engine: &'a Engine,
 }
 
 impl MatchCx<'_> {
     pub fn global(&self, expected: &Global, actual: &crate::Global) -> Result<()> {
-        self.global_ty(expected, actual.wasmtime_ty())
+        self.global_ty(expected, actual.wasmtime_ty(self.store_data))
     }
 
     fn global_ty(&self, expected: &Global, actual: &Global) -> Result<()> {
@@ -28,7 +32,7 @@ impl MatchCx<'_> {
     }
 
     pub fn table(&self, expected: &Table, actual: &crate::Table) -> Result<()> {
-        self.table_ty(expected, actual.wasmtime_ty())
+        self.table_ty(expected, actual.wasmtime_ty(self.store_data))
     }
 
     fn table_ty(&self, expected: &Table, actual: &Table) -> Result<()> {
@@ -50,7 +54,7 @@ impl MatchCx<'_> {
     }
 
     pub fn memory(&self, expected: &Memory, actual: &crate::Memory) -> Result<()> {
-        self.memory_ty(expected, actual.wasmtime_ty())
+        self.memory_ty(expected, actual.wasmtime_ty(self.store_data))
     }
 
     fn memory_ty(&self, expected: &Memory, actual: &Memory) -> Result<()> {
@@ -71,8 +75,24 @@ impl MatchCx<'_> {
     }
 
     pub fn func(&self, expected: SignatureIndex, actual: &crate::Func) -> Result<()> {
+        self.vmshared_signature_index(expected, actual.sig_index(self.store_data))
+    }
+
+    pub(crate) fn host_func(
+        &self,
+        expected: SignatureIndex,
+        actual: &crate::func::HostFunc,
+    ) -> Result<()> {
+        self.vmshared_signature_index(expected, actual.sig_index())
+    }
+
+    pub fn vmshared_signature_index(
+        &self,
+        expected: SignatureIndex,
+        actual: VMSharedSignatureIndex,
+    ) -> Result<()> {
         let matches = match self.signatures.shared_signature(expected) {
-            Some(idx) => actual.sig_index() == idx,
+            Some(idx) => actual == idx,
             // If our expected signature isn't registered, then there's no way
             // that `actual` can match it.
             None => false,
@@ -85,11 +105,11 @@ impl MatchCx<'_> {
     }
 
     pub fn instance(&self, expected: InstanceTypeIndex, actual: &crate::Instance) -> Result<()> {
+        let items = actual.items(self.store_data);
         for (name, expected) in self.types.instance_signatures[expected].exports.iter() {
-            match actual.items.get(name) {
+            match items.get(name) {
                 Some(item) => {
-                    let item = unsafe { Extern::from_wasmtime_export(item, self.store) };
-                    self.extern_(expected, &item)
+                    self.extern_(expected, item)
                         .with_context(|| format!("instance export {:?} incompatible", name))?;
                 }
                 None => bail!("instance type missing export {:?}", name),
@@ -104,7 +124,7 @@ impl MatchCx<'_> {
         // This should only ever be invoked with module linking, and this is an
         // early check that our `field` assertion below should always work as
         // well.
-        assert!(self.store.engine().config().features.module_linking);
+        assert!(self.engine.config().features.module_linking);
 
         let expected_sig = &self.types.module_signatures[expected];
         let module = actual.compiled_module().module();
@@ -149,7 +169,8 @@ impl MatchCx<'_> {
             MatchCx {
                 signatures: actual_signatures,
                 types: actual_types,
-                store: self.store,
+                store_data: self.store_data,
+                engine: self.engine,
             }
             .extern_ty_matches(&actual_ty, expected_ty, self.signatures, self.types)
             .with_context(|| format!("module import {:?} incompatible", name))?;
@@ -292,4 +313,44 @@ impl MatchCx<'_> {
             EntityType::Event(_) => unimplemented!(),
         }
     }
+
+    /// Validates that the `expected` type matches the type of `actual`
+    pub(crate) fn definition(&self, expected: &EntityType, actual: &Definition) -> Result<()> {
+        match actual {
+            Definition::Extern(e) => self.extern_(expected, e),
+            Definition::HostFunc(f) => match expected {
+                EntityType::Function(expected) => self.host_func(*expected, f),
+                _ => bail!("expected {}, but found func", entity_desc(expected)),
+            },
+            Definition::Instance(items) => match expected {
+                EntityType::Instance(expected) => {
+                    for (name, expected) in self.types.instance_signatures[*expected].exports.iter()
+                    {
+                        match items.get(name) {
+                            Some(item) => {
+                                self.definition(expected, item).with_context(|| {
+                                    format!("instance export {:?} incompatible", name)
+                                })?;
+                            }
+                            None => bail!("instance type missing export {:?}", name),
+                        }
+                    }
+                    Ok(())
+                }
+                _ => bail!("expected {}, but found instance", entity_desc(expected)),
+            },
+        }
+    }
+}
+
+fn entity_desc(ty: &EntityType) -> &'static str {
+    match ty {
+        EntityType::Global(_) => "global",
+        EntityType::Table(_) => "table",
+        EntityType::Memory(_) => "memory",
+        EntityType::Function(_) => "func",
+        EntityType::Instance(_) => "instance",
+        EntityType::Module(_) => "module",
+        EntityType::Event(_) => "event",
+    }
 }
